/*-
 * Copyright (c) 2019 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Implementation inspired by Varnish cache_gzip.c
 */

/* for strdup() */
//#define _POSIX_C_SOURCE 200809L

#include "config.h"

#include <stdlib.h>
#include <string.h>
#include <brotli/encode.h>
#include <brotli/decode.h>

#include "cache/cache.h"
#include "cache/cache_filter.h"

#include "vcc_if.h"

/* XXX make this configurable; cf. varnishd default gzip_buffer */
#define BUFFER_SZ (32 * 1024 * 1024)

struct vbr_stream {
	const uint8_t		*next_in;
	uint8_t			*next_out;
	size_t			avail_in;
	size_t			avail_out;
	size_t			total_out;
};

enum vbr_which {ENC, DEC};

struct vbr {
	unsigned		magic;
#define VBR_MAGIC		0x467c387b
	struct vbr_stream	stream;
	union {
		BrotliEncoderState *enc;
		BrotliDecoderState *dec;
	}			state;
	uint8_t			*buf;
	size_t			bufsz;
	size_t			buflen;
	enum vbr_which		which;
};

static struct vbr *
newVBR(enum vbr_which which)
{
	struct vbr *vbr;

	/* XXX inc counters */
	ALLOC_OBJ(vbr, VBR_MAGIC);
	if (vbr == NULL)
		return (NULL);
	if (which == ENC) {
		vbr->state.enc = BrotliEncoderCreateInstance(NULL, NULL, NULL);
		if (vbr->state.enc == NULL)
			return (NULL);
	}
	else {
		vbr->state.dec = BrotliDecoderCreateInstance(NULL, NULL, NULL);
		if (vbr->state.dec == NULL)
			return (NULL);
	}
	/* XXX configure the handle */
	vbr->which = which;
	return (vbr);
}

static void
destroy(struct vbr **vp)
{
	struct vbr *vbr;

	TAKE_OBJ_NOTNULL(vbr, vp, VBR_MAGIC);
	/* XXX log results: bytes in, bytes out */
	/* XXX IMO no finish op necessary here */
	if (vbr->which == ENC)
		BrotliEncoderDestroyInstance(vbr->state.enc);
	else
		BrotliDecoderDestroyInstance(vbr->state.dec);
	if (vbr->buf != NULL)
		free(vbr->buf);
	FREE_OBJ(vbr);
}

static int
getBuf(struct vbr *vbr)
{
	CHECK_OBJ_NOTNULL(vbr, VBR_MAGIC);
	AZ(vbr->bufsz);
	AZ(vbr->buflen);
	AZ(vbr->buf);

	/* XXX make this configurable */
	vbr->bufsz = BUFFER_SZ;
	vbr->buf = malloc(vbr->bufsz);
	if (vbr->buf == NULL) {
		vbr->bufsz = 0;
		return (-1);
	}
	return (0);
}

static const char *
decodeErrMsg(struct vbr *vbr)
{
	CHECK_OBJ_NOTNULL(vbr, VBR_MAGIC);
	assert(vbr->which == DEC);

	return (BrotliDecoderErrorString(
			BrotliDecoderGetErrorCode(vbr->state.dec)));
}

static void
setInputBuf(struct vbr *vbr, const void *ptr, ssize_t len)
{
	CHECK_OBJ_NOTNULL(vbr, VBR_MAGIC);

	AZ(vbr->stream.avail_in);
	vbr->stream.next_in = ptr;
	vbr->stream.avail_in = len;
}

static int
isInputBufEmpty(const struct vbr *vbr)
{
	CHECK_OBJ_NOTNULL(vbr, VBR_MAGIC);
	return (vbr->stream.avail_in == 0);
}

static void
setOutputBuf(struct vbr *vbr, const void *ptr, ssize_t len)
{
	CHECK_OBJ_NOTNULL(vbr, VBR_MAGIC);

	vbr->stream.next_out = TRUST_ME(ptr);
	vbr->stream.avail_out = len;
}

static BROTLI_BOOL
encode(struct vbr *vbr, ssize_t *dl, int finished)
{
	BROTLI_BOOL ret;
	const uint8_t *before;
	enum BrotliEncoderOperation op = BROTLI_OPERATION_PROCESS;

	CHECK_OBJ_NOTNULL(vbr, VBR_MAGIC);
	assert(vbr->which == ENC);
	before = vbr->stream.next_out;

	if (finished)
		op = BROTLI_OPERATION_FINISH;
	ret = BrotliEncoderCompressStream(vbr->state.enc, op,
					  &vbr->stream.avail_in,
					  &vbr->stream.next_in,
					  &vbr->stream.avail_out,
					  &vbr->stream.next_out,
					  &vbr->stream.total_out);
	*dl = (const uint8_t *)vbr->stream.next_out - before;
	return (ret);
}

static BrotliDecoderResult
decode(struct vbr *vbr, ssize_t *dl)
{
	BrotliDecoderResult ret;
	const uint8_t *before;

	CHECK_OBJ_NOTNULL(vbr, VBR_MAGIC);
	assert(vbr->which == DEC);
	before = vbr->stream.next_out;

	ret = BrotliDecoderDecompressStream(vbr->state.dec,
					  &vbr->stream.avail_in,
					  &vbr->stream.next_in,
					  &vbr->stream.avail_out,
					  &vbr->stream.next_out,
					  &vbr->stream.total_out);
	*dl = (const uint8_t *)vbr->stream.next_out - before;
	return (ret);
}

/* VFP interfaces */

static const struct vfp vfp_br, vfp_unbr;

/* init and fini methods for both br and unbr */

static enum vfp_status v_matchproto_(vfp_init_f)
vfp_br_init(struct vfp_ctx *ctx, struct vfp_entry *ent)
{
	struct vbr *vbr;

	CHECK_OBJ_NOTNULL(ctx, VFP_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(ent, VFP_ENTRY_MAGIC);

	/*
	 * Ignore partial responses to range requests (as Varnish does for
	 * gzip and gunzip).
	 */
	if (http_GetStatus(ctx->resp) == 206)
	 	return (VFP_NULL);

	if (ent->vfp == &vfp_br) {
		if (http_GetHdr(ctx->resp, H_Content_Encoding, NULL))
			return (VFP_NULL);
		vbr = newVBR(ENC);
	}
	else {
		if (!http_HdrIs(ctx->resp, H_Content_Encoding, "br"))
			return (VFP_NULL);
		vbr = newVBR(DEC);
	}
	if (vbr == NULL)
		return (VFP_ERROR);
	ent->priv1 = vbr;
	if (getBuf(vbr))
		return (VFP_ERROR);
	setInputBuf(vbr, vbr->buf, 0);
	AZ(vbr->buflen);

	http_Unset(ctx->resp, H_Content_Encoding);
	http_Unset(ctx->resp, H_Content_Length);
	RFC2616_Weaken_Etag(ctx->resp);

	if (ent->vfp == &vfp_br) {
		http_SetHeader(ctx->resp, "Content-Encoding: br");
		RFC2616_Vary_AE(ctx->resp);
	}

	return (VFP_OK);
}

static void v_matchproto_(vfp_fini_f)
vfp_br_fini(struct vfp_ctx *ctx, struct vfp_entry *ent)
{
	struct vbr *vbr;

	CHECK_OBJ_NOTNULL(ctx, VFP_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(ent, VFP_ENTRY_MAGIC);

	if (ent->priv1 != NULL) {
		CAST_OBJ(vbr, ent->priv1, VBR_MAGIC);
		ent->priv1 = NULL;
		destroy(&vbr);
	}
}

/* pull for br (compression) */

static enum vfp_status v_matchproto_(vfp_pull_f)
vfp_br_pull(struct vfp_ctx *ctx, struct vfp_entry *ent, void *ptr,
	    ssize_t *lenp)
{
	struct vbr *vbr;
	ssize_t len, dl;
	enum vfp_status vp = VFP_ERROR;
	int finished = 0;
	BROTLI_BOOL done;

	CHECK_OBJ_NOTNULL(ctx, VFP_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(ent, VFP_ENTRY_MAGIC);
	CAST_OBJ_NOTNULL(vbr, ent->priv1, VBR_MAGIC);
	assert(vbr->which == ENC);
	AN(ptr);
	AN(lenp);

	len = *lenp;
	*lenp = 0;
	setOutputBuf(vbr, ptr, len);

	do {
		if (isInputBufEmpty(vbr)) {
			len = vbr->bufsz;
			vp = VFP_Suck(ctx, vbr->buf, &len);
			if (vp == VFP_ERROR)
				break;
			if (vp == VFP_END)
				finished = 1;
			setInputBuf(vbr, vbr->buf, len);
		}
		if (!isInputBufEmpty(vbr) || finished) {
			done = encode(vbr, &dl, finished);
			if (done != BROTLI_TRUE)
				return (VFP_Error(ctx, "brotli encode failed"));
			if (dl > 0) {
				*lenp = dl;
				return (VFP_OK);
			}
		}
		AN(isInputBufEmpty(vbr));
	} while (!finished);

	if (done != BROTLI_TRUE)
		return (VFP_Error(ctx, "brotli encode failed"));
	return (VFP_END);
}

/* pull for unbr (decompression) */

static enum vfp_status v_matchproto_(vfp_pull_f)
vfp_unbr_pull(struct vfp_ctx *ctx, struct vfp_entry *ent, void *ptr,
	      ssize_t *lenp)
{
	struct vbr *vbr;
	ssize_t len, dl;
	enum vfp_status vp = VFP_ERROR;
	BrotliDecoderResult result = BROTLI_DECODER_RESULT_ERROR;

	CHECK_OBJ_NOTNULL(ctx, VFP_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(ent, VFP_ENTRY_MAGIC);
	CAST_OBJ_NOTNULL(vbr, ent->priv1, VBR_MAGIC);
	assert(vbr->which == DEC);
	AN(ptr);
	AN(lenp);

	len = *lenp;
	*lenp = 0;
	setOutputBuf(vbr, ptr, len);

	do {
		if (isInputBufEmpty(vbr)) {
			len = vbr->bufsz;
			vp = VFP_Suck(ctx, vbr->buf, &len);
			if (vp == VFP_ERROR)
				break;
			setInputBuf(vbr, vbr->buf, len);
		}
		if (!isInputBufEmpty(vbr) || vp == VFP_END) {
			result = decode(vbr, &dl);
			if (result == BROTLI_DECODER_RESULT_SUCCESS &&
			    !isInputBufEmpty(vbr))
				return(VFP_Error(ctx,
						 "Junk after brotli data"));
			if (result == BROTLI_DECODER_RESULT_ERROR)
				return (VFP_Error(ctx,
						  "Invalid brotli data: %s",
						  decodeErrMsg(vbr)));
			if (dl > 0) {
				*lenp = dl;
				return (VFP_OK);
			}
		}
		AN(isInputBufEmpty(vbr));
	} while (result == BROTLI_DECODER_RESULT_NEEDS_MORE_INPUT);

	if (result != BROTLI_DECODER_RESULT_SUCCESS)
		return (VFP_Error(ctx, "Invalid brotli data at end: %s",
				  decodeErrMsg(vbr)));
	return (VFP_END);
}

static const struct vfp vfp_br = {
	.name = "br",
	.init = vfp_br_init,
	.pull = vfp_br_pull,
	.fini = vfp_br_fini,
};

static const struct vfp vfp_unbr = {
	.name = "unbr",
	.init = vfp_br_init,
	.pull = vfp_unbr_pull,
	.fini = vfp_br_fini,
};

/* Event function */

int
vmod_event(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e e)
{
	ASSERT_CLI();
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(priv);

	switch(e) {
	case VCL_EVENT_LOAD:
		VRT_AddVFP(ctx, &vfp_br);
		VRT_AddVFP(ctx, &vfp_unbr);
		return (0);
	case VCL_EVENT_DISCARD:
		VRT_RemoveVFP(ctx, &vfp_br);
		VRT_RemoveVFP(ctx, &vfp_unbr);
		return (0);
	case VCL_EVENT_WARM:
	case VCL_EVENT_COLD:
		return (0);
	default:
		WRONG("illegal event enum");
	}
	NEEDLESS(return (0));
}

static VCL_STRING
brotli_version(VRT_CTX, uint32_t version)
{
	VCL_STRING vers;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	unsigned major = version >> 24;
	unsigned minor = (version >> 12) & 0xFFF;
	unsigned patch = version & 0xFFF;
	if ((vers = WS_Printf(ctx->ws, "%u.%u.%u", major, minor, patch))
	    == NULL) {
		VSLb(ctx->vsl, SLT_VCL_Error, "vfp brotli error: "
		     "insufficient workspace for version string");
		return ("(insufficient workspace for version string)");
	}
	return vers;
}

VCL_STRING
vmod_encoder_version(VRT_CTX)
{
	return (brotli_version(ctx, BrotliEncoderVersion()));
}

VCL_STRING
vmod_decoder_version(VRT_CTX)
{
	return (brotli_version(ctx, BrotliDecoderVersion()));
}

VCL_STRING
vmod_version(VRT_CTX)
{
	(void) ctx;
	return VERSION;
}
